from setuptools import setup
from codecs import open

setup(
    name='threshold_elgamal',
    version='0.0.8',
    long_description_content_type="text/markdown",
    long_description=open('README.md', 'r').read(),
    packages=['tests', 'threshold_elgamal'],
    url='',
    license='',
    author='Katarina Boras',
    author_email='katarina.boras2@outlook.com',
    description='Threshold ElGamal cryptosystem.'
)
